// Code generated by qtc from "alertmanager_request.qtpl". DO NOT EDIT.
// See https://github.com/valyala/quicktemplate for details.

//line app/vmalert/notifier/alertmanager_request.qtpl:1
package notifier

//line app/vmalert/notifier/alertmanager_request.qtpl:1
import (
	"time"

	"github.com/VictoriaMetrics/VictoriaMetrics/lib/prompb"
)

//line app/vmalert/notifier/alertmanager_request.qtpl:8
import (
	qtio422016 "io"

	qt422016 "github.com/valyala/quicktemplate"
)

//line app/vmalert/notifier/alertmanager_request.qtpl:8
var (
	_ = qtio422016.Copy
	_ = qt422016.AcquireByteBuffer
)

//line app/vmalert/notifier/alertmanager_request.qtpl:8
func streamamRequest(qw422016 *qt422016.Writer, alerts []Alert, generatorURL func(Alert) string, lblss [][]prompb.Label) {
//line app/vmalert/notifier/alertmanager_request.qtpl:8
	qw422016.N().S(`[`)
//line app/vmalert/notifier/alertmanager_request.qtpl:10
	for i, alert := range alerts {
//line app/vmalert/notifier/alertmanager_request.qtpl:11
		lbls := lblss[i]

//line app/vmalert/notifier/alertmanager_request.qtpl:11
		qw422016.N().S(`{"startsAt":`)
//line app/vmalert/notifier/alertmanager_request.qtpl:13
		qw422016.N().Q(alert.Start.Format(time.RFC3339Nano))
//line app/vmalert/notifier/alertmanager_request.qtpl:13
		qw422016.N().S(`,"generatorURL":`)
//line app/vmalert/notifier/alertmanager_request.qtpl:14
		qw422016.N().Q(generatorURL(alert))
//line app/vmalert/notifier/alertmanager_request.qtpl:14
		qw422016.N().S(`,`)
//line app/vmalert/notifier/alertmanager_request.qtpl:15
		if !alert.End.IsZero() {
//line app/vmalert/notifier/alertmanager_request.qtpl:15
			qw422016.N().S(`"endsAt":`)
//line app/vmalert/notifier/alertmanager_request.qtpl:16
			qw422016.N().Q(alert.End.Format(time.RFC3339Nano))
//line app/vmalert/notifier/alertmanager_request.qtpl:16
			qw422016.N().S(`,`)
//line app/vmalert/notifier/alertmanager_request.qtpl:17
		}
//line app/vmalert/notifier/alertmanager_request.qtpl:17
		qw422016.N().S(`"labels": {`)
//line app/vmalert/notifier/alertmanager_request.qtpl:19
		ll := len(lbls)

//line app/vmalert/notifier/alertmanager_request.qtpl:20
		for idx, l := range lbls {
//line app/vmalert/notifier/alertmanager_request.qtpl:21
			qw422016.N().Q(l.Name)
//line app/vmalert/notifier/alertmanager_request.qtpl:21
			qw422016.N().S(`:`)
//line app/vmalert/notifier/alertmanager_request.qtpl:21
			qw422016.N().Q(l.Value)
//line app/vmalert/notifier/alertmanager_request.qtpl:21
			if idx != ll-1 {
//line app/vmalert/notifier/alertmanager_request.qtpl:21
				qw422016.N().S(`,`)
//line app/vmalert/notifier/alertmanager_request.qtpl:21
			}
//line app/vmalert/notifier/alertmanager_request.qtpl:22
		}
//line app/vmalert/notifier/alertmanager_request.qtpl:22
		qw422016.N().S(`},"annotations": {`)
//line app/vmalert/notifier/alertmanager_request.qtpl:25
		c := len(alert.Annotations)

//line app/vmalert/notifier/alertmanager_request.qtpl:26
		for k, v := range alert.Annotations {
//line app/vmalert/notifier/alertmanager_request.qtpl:27
			c = c - 1

//line app/vmalert/notifier/alertmanager_request.qtpl:28
			qw422016.N().Q(k)
//line app/vmalert/notifier/alertmanager_request.qtpl:28
			qw422016.N().S(`:`)
//line app/vmalert/notifier/alertmanager_request.qtpl:28
			qw422016.N().Q(v)
//line app/vmalert/notifier/alertmanager_request.qtpl:28
			if c > 0 {
//line app/vmalert/notifier/alertmanager_request.qtpl:28
				qw422016.N().S(`,`)
//line app/vmalert/notifier/alertmanager_request.qtpl:28
			}
//line app/vmalert/notifier/alertmanager_request.qtpl:29
		}
//line app/vmalert/notifier/alertmanager_request.qtpl:29
		qw422016.N().S(`}}`)
//line app/vmalert/notifier/alertmanager_request.qtpl:32
		if i != len(alerts)-1 {
//line app/vmalert/notifier/alertmanager_request.qtpl:32
			qw422016.N().S(`,`)
//line app/vmalert/notifier/alertmanager_request.qtpl:32
		}
//line app/vmalert/notifier/alertmanager_request.qtpl:33
	}
//line app/vmalert/notifier/alertmanager_request.qtpl:33
	qw422016.N().S(`]`)
//line app/vmalert/notifier/alertmanager_request.qtpl:35
}

//line app/vmalert/notifier/alertmanager_request.qtpl:35
func writeamRequest(qq422016 qtio422016.Writer, alerts []Alert, generatorURL func(Alert) string, lblss [][]prompb.Label) {
//line app/vmalert/notifier/alertmanager_request.qtpl:35
	qw422016 := qt422016.AcquireWriter(qq422016)
//line app/vmalert/notifier/alertmanager_request.qtpl:35
	streamamRequest(qw422016, alerts, generatorURL, lblss)
//line app/vmalert/notifier/alertmanager_request.qtpl:35
	qt422016.ReleaseWriter(qw422016)
//line app/vmalert/notifier/alertmanager_request.qtpl:35
}

//line app/vmalert/notifier/alertmanager_request.qtpl:35
func amRequest(alerts []Alert, generatorURL func(Alert) string, lblss [][]prompb.Label) string {
//line app/vmalert/notifier/alertmanager_request.qtpl:35
	qb422016 := qt422016.AcquireByteBuffer()
//line app/vmalert/notifier/alertmanager_request.qtpl:35
	writeamRequest(qb422016, alerts, generatorURL, lblss)
//line app/vmalert/notifier/alertmanager_request.qtpl:35
	qs422016 := string(qb422016.B)
//line app/vmalert/notifier/alertmanager_request.qtpl:35
	qt422016.ReleaseByteBuffer(qb422016)
//line app/vmalert/notifier/alertmanager_request.qtpl:35
	return qs422016
//line app/vmalert/notifier/alertmanager_request.qtpl:35
}
