// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type StringSliceWrapper struct {
	orig  *[]string
	state *State
}

func GetStringSliceOrig(ms StringSliceWrapper) *[]string {
	return ms.orig
}

func GetStringSliceState(ms StringSliceWrapper) *State {
	return ms.state
}

func NewStringSliceWrapper(orig *[]string, state *State) StringSliceWrapper {
	return StringSliceWrapper{orig: orig, state: state}
}

func GenTestStringSliceWrapper() StringSliceWrapper {
	orig := []string{"a", "b", "c"}
	return NewStringSliceWrapper(&orig, NewState())
}

func GenTestStringSlice() []string {
	return []string{"a", "b", "c"}
}
