// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type ResourceWrapper struct {
	orig  *Resource
	state *State
}

func GetResourceOrig(ms ResourceWrapper) *Resource {
	return ms.orig
}

func GetResourceState(ms ResourceWrapper) *State {
	return ms.state
}

func NewResourceWrapper(orig *Resource, state *State) ResourceWrapper {
	return ResourceWrapper{orig: orig, state: state}
}

func GenTestResourceWrapper() ResourceWrapper {
	return NewResourceWrapper(GenTestResource(), NewState())
}
