// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type UInt64SliceWrapper struct {
	orig  *[]uint64
	state *State
}

func GetUInt64SliceOrig(ms UInt64SliceWrapper) *[]uint64 {
	return ms.orig
}

func GetUInt64SliceState(ms UInt64SliceWrapper) *State {
	return ms.state
}

func NewUInt64SliceWrapper(orig *[]uint64, state *State) UInt64SliceWrapper {
	return UInt64SliceWrapper{orig: orig, state: state}
}

func GenTestUInt64SliceWrapper() UInt64SliceWrapper {
	orig := []uint64{1, 2, 3}
	return NewUInt64SliceWrapper(&orig, NewState())
}

func GenTestUint64Slice() []uint64 {
	return []uint64{1, 2, 3}
}
