// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type EntityRefSliceWrapper struct {
	orig  *[]*EntityRef
	state *State
}

func GetEntityRefSliceOrig(ms EntityRefSliceWrapper) *[]*EntityRef {
	return ms.orig
}

func GetEntityRefSliceState(ms EntityRefSliceWrapper) *State {
	return ms.state
}

func NewEntityRefSliceWrapper(orig *[]*EntityRef, state *State) EntityRefSliceWrapper {
	return EntityRefSliceWrapper{orig: orig, state: state}
}

func GenTestEntityRefSliceWrapper() EntityRefSliceWrapper {
	orig := GenTestEntityRefPtrSlice()
	return NewEntityRefSliceWrapper(&orig, NewState())
}
