// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type Int32SliceWrapper struct {
	orig  *[]int32
	state *State
}

func GetInt32SliceOrig(ms Int32SliceWrapper) *[]int32 {
	return ms.orig
}

func GetInt32SliceState(ms Int32SliceWrapper) *State {
	return ms.state
}

func NewInt32SliceWrapper(orig *[]int32, state *State) Int32SliceWrapper {
	return Int32SliceWrapper{orig: orig, state: state}
}

func GenTestInt32SliceWrapper() Int32SliceWrapper {
	orig := []int32{1, 2, 3}
	return NewInt32SliceWrapper(&orig, NewState())
}

func GenTestInt32Slice() []int32 {
	return []int32{1, 2, 3}
}
