// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type EntityRefWrapper struct {
	orig  *EntityRef
	state *State
}

func GetEntityRefOrig(ms EntityRefWrapper) *EntityRef {
	return ms.orig
}

func GetEntityRefState(ms EntityRefWrapper) *State {
	return ms.state
}

func NewEntityRefWrapper(orig *EntityRef, state *State) EntityRefWrapper {
	return EntityRefWrapper{orig: orig, state: state}
}

func GenTestEntityRefWrapper() EntityRefWrapper {
	return NewEntityRefWrapper(GenTestEntityRef(), NewState())
}
