// Code generated by qtc from "tags_response.qtpl". DO NOT EDIT.
// See https://github.com/valyala/quicktemplate for details.

// Tags generates response for /tags handlerSee https://graphite.readthedocs.io/en/stable/tags.html#exploring-tags

//line app/vmselect/graphite/tags_response.qtpl:5
package graphite

//line app/vmselect/graphite/tags_response.qtpl:5
import (
	qtio422016 "io"

	qt422016 "github.com/valyala/quicktemplate"
)

//line app/vmselect/graphite/tags_response.qtpl:5
var (
	_ = qtio422016.Copy
	_ = qt422016.AcquireByteBuffer
)

//line app/vmselect/graphite/tags_response.qtpl:5
func StreamTagsResponse(qw422016 *qt422016.Writer, tags []string) {
//line app/vmselect/graphite/tags_response.qtpl:5
	qw422016.N().S(`[`)
//line app/vmselect/graphite/tags_response.qtpl:7
	for i, tag := range tags {
//line app/vmselect/graphite/tags_response.qtpl:7
		qw422016.N().S(`{"tag":`)
//line app/vmselect/graphite/tags_response.qtpl:9
		qw422016.N().Q(tag)
//line app/vmselect/graphite/tags_response.qtpl:9
		qw422016.N().S(`}`)
//line app/vmselect/graphite/tags_response.qtpl:11
		if i+1 < len(tags) {
//line app/vmselect/graphite/tags_response.qtpl:11
			qw422016.N().S(`,`)
//line app/vmselect/graphite/tags_response.qtpl:11
		}
//line app/vmselect/graphite/tags_response.qtpl:12
	}
//line app/vmselect/graphite/tags_response.qtpl:12
	qw422016.N().S(`]`)
//line app/vmselect/graphite/tags_response.qtpl:14
}

//line app/vmselect/graphite/tags_response.qtpl:14
func WriteTagsResponse(qq422016 qtio422016.Writer, tags []string) {
//line app/vmselect/graphite/tags_response.qtpl:14
	qw422016 := qt422016.AcquireWriter(qq422016)
//line app/vmselect/graphite/tags_response.qtpl:14
	StreamTagsResponse(qw422016, tags)
//line app/vmselect/graphite/tags_response.qtpl:14
	qt422016.ReleaseWriter(qw422016)
//line app/vmselect/graphite/tags_response.qtpl:14
}

//line app/vmselect/graphite/tags_response.qtpl:14
func TagsResponse(tags []string) string {
//line app/vmselect/graphite/tags_response.qtpl:14
	qb422016 := qt422016.AcquireByteBuffer()
//line app/vmselect/graphite/tags_response.qtpl:14
	WriteTagsResponse(qb422016, tags)
//line app/vmselect/graphite/tags_response.qtpl:14
	qs422016 := string(qb422016.B)
//line app/vmselect/graphite/tags_response.qtpl:14
	qt422016.ReleaseByteBuffer(qb422016)
//line app/vmselect/graphite/tags_response.qtpl:14
	return qs422016
//line app/vmselect/graphite/tags_response.qtpl:14
}
