{{- define "test.1" -}}
    {{ printf "Hello %s!" externalURL }}
{{- end -}}
{{- define "test.0" -}}
    {{ printf "Hello %s!" externalURL }}
{{- end -}}
{{- define "test.3" -}}
    {{ printf "Hello %s!" externalURL }}
{{- end -}}