// Code generated by qtc from "tags_find_series_response.qtpl". DO NOT EDIT.
// See https://github.com/valyala/quicktemplate for details.

//line app/vmselect/graphite/tags_find_series_response.qtpl:3
package graphite

//line app/vmselect/graphite/tags_find_series_response.qtpl:3
import (
	qtio422016 "io"

	qt422016 "github.com/valyala/quicktemplate"
)

//line app/vmselect/graphite/tags_find_series_response.qtpl:3
var (
	_ = qtio422016.Copy
	_ = qt422016.AcquireByteBuffer
)

//line app/vmselect/graphite/tags_find_series_response.qtpl:3
func StreamTagsFindSeriesResponse(qw422016 *qt422016.Writer, paths []string) {
//line app/vmselect/graphite/tags_find_series_response.qtpl:3
	qw422016.N().S(`[`)
//line app/vmselect/graphite/tags_find_series_response.qtpl:5
	for i, path := range paths {
//line app/vmselect/graphite/tags_find_series_response.qtpl:6
		qw422016.N().Q(path)
//line app/vmselect/graphite/tags_find_series_response.qtpl:7
		if i+1 < len(paths) {
//line app/vmselect/graphite/tags_find_series_response.qtpl:7
			qw422016.N().S(`,`)
//line app/vmselect/graphite/tags_find_series_response.qtpl:7
		}
//line app/vmselect/graphite/tags_find_series_response.qtpl:8
	}
//line app/vmselect/graphite/tags_find_series_response.qtpl:8
	qw422016.N().S(`]`)
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
}

//line app/vmselect/graphite/tags_find_series_response.qtpl:10
func WriteTagsFindSeriesResponse(qq422016 qtio422016.Writer, paths []string) {
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
	qw422016 := qt422016.AcquireWriter(qq422016)
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
	StreamTagsFindSeriesResponse(qw422016, paths)
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
	qt422016.ReleaseWriter(qw422016)
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
}

//line app/vmselect/graphite/tags_find_series_response.qtpl:10
func TagsFindSeriesResponse(paths []string) string {
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
	qb422016 := qt422016.AcquireByteBuffer()
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
	WriteTagsFindSeriesResponse(qb422016, paths)
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
	qs422016 := string(qb422016.B)
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
	qt422016.ReleaseByteBuffer(qb422016)
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
	return qs422016
//line app/vmselect/graphite/tags_find_series_response.qtpl:10
}
