// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type ProfilesWrapper struct {
	orig  *ExportProfilesServiceRequest
	state *State
}

func GetProfilesOrig(ms ProfilesWrapper) *ExportProfilesServiceRequest {
	return ms.orig
}

func GetProfilesState(ms ProfilesWrapper) *State {
	return ms.state
}

func NewProfilesWrapper(orig *ExportProfilesServiceRequest, state *State) ProfilesWrapper {
	return ProfilesWrapper{orig: orig, state: state}
}

func GenTestProfilesWrapper() ProfilesWrapper {
	return NewProfilesWrapper(GenTestExportProfilesServiceRequest(), NewState())
}
