// Code generated by mdatagen. DO NOT EDIT.

package metadata

import (
	"go.opentelemetry.io/collector/featuregate"
)

var PdataUseCustomProtoEncodingFeatureGate = featuregate.GlobalRegistry().MustRegister(
	"pdata.useCustomProtoEncoding",
	featuregate.StageStable,
	featuregate.WithRegisterDescription("When enabled, enable custom proto encoding. This is a required step to enable featuregate pdata.useProtoPooling."),
	featuregate.WithRegisterReferenceURL("https://github.com/open-telemetry/opentelemetry-collector/issues/13631"),
	featuregate.WithRegisterFromVersion("v0.133.0"),
	featuregate.WithRegisterToVersion("v0.137.0"),
)

var PdataUseProtoPoolingFeatureGate = featuregate.GlobalRegistry().MustRegister(
	"pdata.useProtoPooling",
	featuregate.StageAlpha,
	featuregate.WithRegisterDescription("When enabled, enable using local memory pools for underlying data that the pdata messages are pushed to."),
	featuregate.WithRegisterReferenceURL("https://github.com/open-telemetry/opentelemetry-collector/issues/13631"),
	featuregate.WithRegisterFromVersion("v0.133.0"),
)
