package mergeset

const (
	metaindexFilename = "metaindex.bin"
	indexFilename     = "index.bin"
	itemsFilename     = "items.bin"
	lensFilename      = "lens.bin"
	metadataFilename  = "metadata.json"
	partsFilename     = "parts.json"
)
