// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type MetricsWrapper struct {
	orig  *ExportMetricsServiceRequest
	state *State
}

func GetMetricsOrig(ms MetricsWrapper) *ExportMetricsServiceRequest {
	return ms.orig
}

func GetMetricsState(ms MetricsWrapper) *State {
	return ms.state
}

func NewMetricsWrapper(orig *ExportMetricsServiceRequest, state *State) MetricsWrapper {
	return MetricsWrapper{orig: orig, state: state}
}

func GenTestMetricsWrapper() MetricsWrapper {
	return NewMetricsWrapper(GenTestExportMetricsServiceRequest(), NewState())
}
