// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type ByteSliceWrapper struct {
	orig  *[]byte
	state *State
}

func GetByteSliceOrig(ms ByteSliceWrapper) *[]byte {
	return ms.orig
}

func GetByteSliceState(ms ByteSliceWrapper) *State {
	return ms.state
}

func NewByteSliceWrapper(orig *[]byte, state *State) ByteSliceWrapper {
	return ByteSliceWrapper{orig: orig, state: state}
}

func GenTestByteSliceWrapper() ByteSliceWrapper {
	orig := []byte{1, 2, 3}
	return NewByteSliceWrapper(&orig, NewState())
}

func GenTestByteSlice() []byte {
	return []byte{1, 2, 3}
}
