// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type TracesWrapper struct {
	orig  *ExportTraceServiceRequest
	state *State
}

func GetTracesOrig(ms TracesWrapper) *ExportTraceServiceRequest {
	return ms.orig
}

func GetTracesState(ms TracesWrapper) *State {
	return ms.state
}

func NewTracesWrapper(orig *ExportTraceServiceRequest, state *State) TracesWrapper {
	return TracesWrapper{orig: orig, state: state}
}

func GenTestTracesWrapper() TracesWrapper {
	return NewTracesWrapper(GenTestExportTraceServiceRequest(), NewState())
}
