// Code generated by qtc from "footer.qtpl". DO NOT EDIT.
// See https://github.com/valyala/quicktemplate for details.

//line app/vmalert/tpl/footer.qtpl:1
package tpl

//line app/vmalert/tpl/footer.qtpl:1
import (
	"net/http"
)

//line app/vmalert/tpl/footer.qtpl:6
import (
	qtio422016 "io"

	qt422016 "github.com/valyala/quicktemplate"
)

//line app/vmalert/tpl/footer.qtpl:6
var (
	_ = qtio422016.Copy
	_ = qt422016.AcquireByteBuffer
)

//line app/vmalert/tpl/footer.qtpl:6
func StreamFooter(qw422016 *qt422016.Writer, r *http.Request) {
//line app/vmalert/tpl/footer.qtpl:6
	qw422016.N().S(`
        </main>
    </body>
</html>
`)
//line app/vmalert/tpl/footer.qtpl:10
}

//line app/vmalert/tpl/footer.qtpl:10
func WriteFooter(qq422016 qtio422016.Writer, r *http.Request) {
//line app/vmalert/tpl/footer.qtpl:10
	qw422016 := qt422016.AcquireWriter(qq422016)
//line app/vmalert/tpl/footer.qtpl:10
	StreamFooter(qw422016, r)
//line app/vmalert/tpl/footer.qtpl:10
	qt422016.ReleaseWriter(qw422016)
//line app/vmalert/tpl/footer.qtpl:10
}

//line app/vmalert/tpl/footer.qtpl:10
func Footer(r *http.Request) string {
//line app/vmalert/tpl/footer.qtpl:10
	qb422016 := qt422016.AcquireByteBuffer()
//line app/vmalert/tpl/footer.qtpl:10
	WriteFooter(qb422016, r)
//line app/vmalert/tpl/footer.qtpl:10
	qs422016 := string(qb422016.B)
//line app/vmalert/tpl/footer.qtpl:10
	qt422016.ReleaseByteBuffer(qb422016)
//line app/vmalert/tpl/footer.qtpl:10
	return qs422016
//line app/vmalert/tpl/footer.qtpl:10
}
