// Code generated by qtc from "metrics_index_response.qtpl". DO NOT EDIT.
// See https://github.com/valyala/quicktemplate for details.

// MetricsIndexResponse generates response for /metrics/index.json .See https://graphite-api.readthedocs.io/en/latest/api.html#metrics-index-json

//line app/vmselect/graphite/metrics_index_response.qtpl:5
package graphite

//line app/vmselect/graphite/metrics_index_response.qtpl:5
import (
	qtio422016 "io"

	qt422016 "github.com/valyala/quicktemplate"
)

//line app/vmselect/graphite/metrics_index_response.qtpl:5
var (
	_ = qtio422016.Copy
	_ = qt422016.AcquireByteBuffer
)

//line app/vmselect/graphite/metrics_index_response.qtpl:5
func StreamMetricsIndexResponse(qw422016 *qt422016.Writer, metricNames []string, jsonp string) {
//line app/vmselect/graphite/metrics_index_response.qtpl:6
	if jsonp != "" {
//line app/vmselect/graphite/metrics_index_response.qtpl:6
		qw422016.N().S(jsonp)
//line app/vmselect/graphite/metrics_index_response.qtpl:6
		qw422016.N().S(`(`)
//line app/vmselect/graphite/metrics_index_response.qtpl:6
	}
//line app/vmselect/graphite/metrics_index_response.qtpl:7
	streammetricPaths(qw422016, metricNames)
//line app/vmselect/graphite/metrics_index_response.qtpl:8
	if jsonp != "" {
//line app/vmselect/graphite/metrics_index_response.qtpl:8
		qw422016.N().S(`)`)
//line app/vmselect/graphite/metrics_index_response.qtpl:8
	}
//line app/vmselect/graphite/metrics_index_response.qtpl:9
}

//line app/vmselect/graphite/metrics_index_response.qtpl:9
func WriteMetricsIndexResponse(qq422016 qtio422016.Writer, metricNames []string, jsonp string) {
//line app/vmselect/graphite/metrics_index_response.qtpl:9
	qw422016 := qt422016.AcquireWriter(qq422016)
//line app/vmselect/graphite/metrics_index_response.qtpl:9
	StreamMetricsIndexResponse(qw422016, metricNames, jsonp)
//line app/vmselect/graphite/metrics_index_response.qtpl:9
	qt422016.ReleaseWriter(qw422016)
//line app/vmselect/graphite/metrics_index_response.qtpl:9
}

//line app/vmselect/graphite/metrics_index_response.qtpl:9
func MetricsIndexResponse(metricNames []string, jsonp string) string {
//line app/vmselect/graphite/metrics_index_response.qtpl:9
	qb422016 := qt422016.AcquireByteBuffer()
//line app/vmselect/graphite/metrics_index_response.qtpl:9
	WriteMetricsIndexResponse(qb422016, metricNames, jsonp)
//line app/vmselect/graphite/metrics_index_response.qtpl:9
	qs422016 := string(qb422016.B)
//line app/vmselect/graphite/metrics_index_response.qtpl:9
	qt422016.ReleaseByteBuffer(qb422016)
//line app/vmselect/graphite/metrics_index_response.qtpl:9
	return qs422016
//line app/vmselect/graphite/metrics_index_response.qtpl:9
}
