// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type InstrumentationScopeWrapper struct {
	orig  *InstrumentationScope
	state *State
}

func GetInstrumentationScopeOrig(ms InstrumentationScopeWrapper) *InstrumentationScope {
	return ms.orig
}

func GetInstrumentationScopeState(ms InstrumentationScopeWrapper) *State {
	return ms.state
}

func NewInstrumentationScopeWrapper(orig *InstrumentationScope, state *State) InstrumentationScopeWrapper {
	return InstrumentationScopeWrapper{orig: orig, state: state}
}

func GenTestInstrumentationScopeWrapper() InstrumentationScopeWrapper {
	return NewInstrumentationScopeWrapper(GenTestInstrumentationScope(), NewState())
}
