// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type LogsWrapper struct {
	orig  *ExportLogsServiceRequest
	state *State
}

func GetLogsOrig(ms LogsWrapper) *ExportLogsServiceRequest {
	return ms.orig
}

func GetLogsState(ms LogsWrapper) *State {
	return ms.state
}

func NewLogsWrapper(orig *ExportLogsServiceRequest, state *State) LogsWrapper {
	return LogsWrapper{orig: orig, state: state}
}

func GenTestLogsWrapper() LogsWrapper {
	return NewLogsWrapper(GenTestExportLogsServiceRequest(), NewState())
}
