// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Package component outlines the components used in the collector
// and provides a foundation for the component’s creation and
// termination process. A component can be either a receiver, exporter,
// processor, an extension, or a connector.
package component // import "go.opentelemetry.io/collector/component"
