// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

const (
	SeverityNumber_SEVERITY_NUMBER_UNSPECIFIED = SeverityNumber(0)
	SeverityNumber_SEVERITY_NUMBER_TRACE       = SeverityNumber(1)
	SeverityNumber_SEVERITY_NUMBER_TRACE2      = SeverityNumber(2)
	SeverityNumber_SEVERITY_NUMBER_TRACE3      = SeverityNumber(3)
	SeverityNumber_SEVERITY_NUMBER_TRACE4      = SeverityNumber(4)
	SeverityNumber_SEVERITY_NUMBER_DEBUG       = SeverityNumber(5)
	SeverityNumber_SEVERITY_NUMBER_DEBUG2      = SeverityNumber(6)
	SeverityNumber_SEVERITY_NUMBER_DEBUG3      = SeverityNumber(7)
	SeverityNumber_SEVERITY_NUMBER_DEBUG4      = SeverityNumber(8)
	SeverityNumber_SEVERITY_NUMBER_INFO        = SeverityNumber(9)
	SeverityNumber_SEVERITY_NUMBER_INFO2       = SeverityNumber(10)
	SeverityNumber_SEVERITY_NUMBER_INFO3       = SeverityNumber(11)
	SeverityNumber_SEVERITY_NUMBER_INFO4       = SeverityNumber(12)
	SeverityNumber_SEVERITY_NUMBER_WARN        = SeverityNumber(13)
	SeverityNumber_SEVERITY_NUMBER_WARN2       = SeverityNumber(14)
	SeverityNumber_SEVERITY_NUMBER_WARN3       = SeverityNumber(15)
	SeverityNumber_SEVERITY_NUMBER_WARN4       = SeverityNumber(16)
	SeverityNumber_SEVERITY_NUMBER_ERROR       = SeverityNumber(17)
	SeverityNumber_SEVERITY_NUMBER_ERROR2      = SeverityNumber(18)
	SeverityNumber_SEVERITY_NUMBER_ERROR3      = SeverityNumber(19)
	SeverityNumber_SEVERITY_NUMBER_ERROR4      = SeverityNumber(20)
	SeverityNumber_SEVERITY_NUMBER_FATAL       = SeverityNumber(21)
	SeverityNumber_SEVERITY_NUMBER_FATAL2      = SeverityNumber(22)
	SeverityNumber_SEVERITY_NUMBER_FATAL3      = SeverityNumber(23)
	SeverityNumber_SEVERITY_NUMBER_FATAL4      = SeverityNumber(24)
)

// SeverityNumber represent possible values for LogRecord.SeverityNumber
type SeverityNumber int32

var SeverityNumber_name = map[int32]string{
	0:  "SEVERITY_NUMBER_UNSPECIFIED",
	1:  "SEVERITY_NUMBER_TRACE ",
	2:  "SEVERITY_NUMBER_TRACE2",
	3:  "SEVERITY_NUMBER_TRACE3",
	4:  "SEVERITY_NUMBER_TRACE4",
	5:  "SEVERITY_NUMBER_DEBUG",
	6:  "SEVERITY_NUMBER_DEBUG2",
	7:  "SEVERITY_NUMBER_DEBUG3",
	8:  "SEVERITY_NUMBER_DEBUG4",
	9:  "SEVERITY_NUMBER_INFO",
	10: "SEVERITY_NUMBER_INFO2",
	11: "SEVERITY_NUMBER_INFO3",
	12: "SEVERITY_NUMBER_INFO4",
	13: "SEVERITY_NUMBER_WARN",
	14: "SEVERITY_NUMBER_WARN2",
	15: "SEVERITY_NUMBER_WARN3",
	16: "SEVERITY_NUMBER_WARN4",
	17: "SEVERITY_NUMBER_ERROR",
	18: "SEVERITY_NUMBER_ERROR2",
	19: "SEVERITY_NUMBER_ERROR3",
	20: "SEVERITY_NUMBER_ERROR4",
	21: "SEVERITY_NUMBER_FATAL",
	22: "SEVERITY_NUMBER_FATAL2",
	23: "SEVERITY_NUMBER_FATAL3",
	24: "SEVERITY_NUMBER_FATAL4",
}

var SeverityNumber_value = map[string]int32{
	"SEVERITY_NUMBER_UNSPECIFIED": 0,
	"SEVERITY_NUMBER_TRACE ":      1,
	"SEVERITY_NUMBER_TRACE2":      2,
	"SEVERITY_NUMBER_TRACE3":      3,
	"SEVERITY_NUMBER_TRACE4":      4,
	"SEVERITY_NUMBER_DEBUG":       5,
	"SEVERITY_NUMBER_DEBUG2":      6,
	"SEVERITY_NUMBER_DEBUG3":      7,
	"SEVERITY_NUMBER_DEBUG4":      8,
	"SEVERITY_NUMBER_INFO":        9,
	"SEVERITY_NUMBER_INFO2":       10,
	"SEVERITY_NUMBER_INFO3":       11,
	"SEVERITY_NUMBER_INFO4":       12,
	"SEVERITY_NUMBER_WARN":        13,
	"SEVERITY_NUMBER_WARN2":       14,
	"SEVERITY_NUMBER_WARN3":       15,
	"SEVERITY_NUMBER_WARN4":       16,
	"SEVERITY_NUMBER_ERROR":       17,
	"SEVERITY_NUMBER_ERROR2":      18,
	"SEVERITY_NUMBER_ERROR3":      19,
	"SEVERITY_NUMBER_ERROR4":      20,
	"SEVERITY_NUMBER_FATAL":       21,
	"SEVERITY_NUMBER_FATAL2":      22,
	"SEVERITY_NUMBER_FATAL3":      23,
	"SEVERITY_NUMBER_FATAL4":      24,
}
