// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type Float64SliceWrapper struct {
	orig  *[]float64
	state *State
}

func GetFloat64SliceOrig(ms Float64SliceWrapper) *[]float64 {
	return ms.orig
}

func GetFloat64SliceState(ms Float64SliceWrapper) *State {
	return ms.state
}

func NewFloat64SliceWrapper(orig *[]float64, state *State) Float64SliceWrapper {
	return Float64SliceWrapper{orig: orig, state: state}
}

func GenTestFloat64SliceWrapper() Float64SliceWrapper {
	orig := []float64{1.1, 2.2, 3.3}
	return NewFloat64SliceWrapper(&orig, NewState())
}

func GenTestFloat64Slice() []float64 {
	return []float64{1.1, 2.2, 3.3}
}
