// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type SliceWrapper struct {
	orig  *[]AnyValue
	state *State
}

func GetSliceOrig(ms SliceWrapper) *[]AnyValue {
	return ms.orig
}

func GetSliceState(ms SliceWrapper) *State {
	return ms.state
}

func NewSliceWrapper(orig *[]AnyValue, state *State) SliceWrapper {
	return SliceWrapper{orig: orig, state: state}
}

func GenTestSliceWrapper() SliceWrapper {
	orig := GenTestAnyValueSlice()
	return NewSliceWrapper(&orig, NewState())
}
