// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type ProfilesDataWrapper struct {
	orig  *ProfilesData
	state *State
}

func GetProfilesDataOrig(ms ProfilesDataWrapper) *ProfilesData {
	return ms.orig
}

func GetProfilesDataState(ms ProfilesDataWrapper) *State {
	return ms.state
}

func NewProfilesDataWrapper(orig *ProfilesData, state *State) ProfilesDataWrapper {
	return ProfilesDataWrapper{orig: orig, state: state}
}

func GenTestProfilesDataWrapper() ProfilesDataWrapper {
	return NewProfilesDataWrapper(GenTestProfilesData(), NewState())
}
