// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

type Int64SliceWrapper struct {
	orig  *[]int64
	state *State
}

func GetInt64SliceOrig(ms Int64SliceWrapper) *[]int64 {
	return ms.orig
}

func GetInt64SliceState(ms Int64SliceWrapper) *State {
	return ms.state
}

func NewInt64SliceWrapper(orig *[]int64, state *State) Int64SliceWrapper {
	return Int64SliceWrapper{orig: orig, state: state}
}

func GenTestInt64SliceWrapper() Int64SliceWrapper {
	orig := []int64{1, 2, 3}
	return NewInt64SliceWrapper(&orig, NewState())
}

func GenTestInt64Slice() []int64 {
	return []int64{1, 2, 3}
}
