// Copyright The OpenTelemetry Authors
// SPDX-License-Identifier: Apache-2.0

// Code generated by "internal/cmd/pdatagen/main.go". DO NOT EDIT.
// To regenerate this file run "make genpdata".

package internal

const (
	SpanKind_SPAN_KIND_UNSPECIFIED = SpanKind(0)
	SpanKind_SPAN_KIND_INTERNAL    = SpanKind(1)
	SpanKind_SPAN_KIND_SERVER      = SpanKind(2)
	SpanKind_SPAN_KIND_CLIENT      = SpanKind(3)
	SpanKind_SPAN_KIND_PRODUCER    = SpanKind(4)
	SpanKind_SPAN_KIND_CONSUMER    = SpanKind(5)
)

// SpanKind is the type of span.
// Can be used to specify additional relationships between spans in addition to a parent/child relationship.
type SpanKind int32

var SpanKind_name = map[int32]string{
	0: "SPAN_KIND_UNSPECIFIED",
	1: "SPAN_KIND_INTERNAL",
	2: "SPAN_KIND_SERVER",
	3: "SPAN_KIND_CLIENT",
	4: "SPAN_KIND_PRODUCER",
	5: "SPAN_KIND_CONSUMER",
}

var SpanKind_value = map[string]int32{
	"SPAN_KIND_UNSPECIFIED": 0,
	"SPAN_KIND_INTERNAL":    1,
	"SPAN_KIND_SERVER":      2,
	"SPAN_KIND_CLIENT":      3,
	"SPAN_KIND_PRODUCER":    4,
	"SPAN_KIND_CONSUMER":    5,
}
