// Code generated by qtc from "series_response.qtpl". DO NOT EDIT.
// See https://github.com/valyala/quicktemplate for details.

//line app/vmselect/prometheus/series_response.qtpl:1
package prometheus

//line app/vmselect/prometheus/series_response.qtpl:1
import (
	"github.com/VictoriaMetrics/VictoriaMetrics/lib/querytracer"
	"github.com/VictoriaMetrics/VictoriaMetrics/lib/storage"
)

// SeriesResponse generates response for /api/v1/series.See https://prometheus.io/docs/prometheus/latest/querying/api/#finding-series-by-label-matchers

//line app/vmselect/prometheus/series_response.qtpl:9
import (
	qtio422016 "io"

	qt422016 "github.com/valyala/quicktemplate"
)

//line app/vmselect/prometheus/series_response.qtpl:9
var (
	_ = qtio422016.Copy
	_ = qt422016.AcquireByteBuffer
)

//line app/vmselect/prometheus/series_response.qtpl:9
func StreamSeriesResponse(qw422016 *qt422016.Writer, metricNames []string, qt *querytracer.Tracer, qtDone func()) {
//line app/vmselect/prometheus/series_response.qtpl:9
	qw422016.N().S(`{"status":"success","data":[`)
//line app/vmselect/prometheus/series_response.qtpl:13
	var mn storage.MetricName

//line app/vmselect/prometheus/series_response.qtpl:14
	for i, metricName := range metricNames {
//line app/vmselect/prometheus/series_response.qtpl:15
		err := mn.UnmarshalString(metricName)

//line app/vmselect/prometheus/series_response.qtpl:16
		if err != nil {
//line app/vmselect/prometheus/series_response.qtpl:17
			qw422016.N().Q(err.Error())
//line app/vmselect/prometheus/series_response.qtpl:18
		} else {
//line app/vmselect/prometheus/series_response.qtpl:19
			streammetricNameObject(qw422016, &mn)
//line app/vmselect/prometheus/series_response.qtpl:20
		}
//line app/vmselect/prometheus/series_response.qtpl:21
		if i+1 < len(metricNames) {
//line app/vmselect/prometheus/series_response.qtpl:21
			qw422016.N().S(`,`)
//line app/vmselect/prometheus/series_response.qtpl:21
		}
//line app/vmselect/prometheus/series_response.qtpl:22
	}
//line app/vmselect/prometheus/series_response.qtpl:22
	qw422016.N().S(`]`)
//line app/vmselect/prometheus/series_response.qtpl:25
	qt.Printf("generate response: series=%d", len(metricNames))
	qtDone()

//line app/vmselect/prometheus/series_response.qtpl:28
	streamdumpQueryTrace(qw422016, qt)
//line app/vmselect/prometheus/series_response.qtpl:28
	qw422016.N().S(`}`)
//line app/vmselect/prometheus/series_response.qtpl:30
}

//line app/vmselect/prometheus/series_response.qtpl:30
func WriteSeriesResponse(qq422016 qtio422016.Writer, metricNames []string, qt *querytracer.Tracer, qtDone func()) {
//line app/vmselect/prometheus/series_response.qtpl:30
	qw422016 := qt422016.AcquireWriter(qq422016)
//line app/vmselect/prometheus/series_response.qtpl:30
	StreamSeriesResponse(qw422016, metricNames, qt, qtDone)
//line app/vmselect/prometheus/series_response.qtpl:30
	qt422016.ReleaseWriter(qw422016)
//line app/vmselect/prometheus/series_response.qtpl:30
}

//line app/vmselect/prometheus/series_response.qtpl:30
func SeriesResponse(metricNames []string, qt *querytracer.Tracer, qtDone func()) string {
//line app/vmselect/prometheus/series_response.qtpl:30
	qb422016 := qt422016.AcquireByteBuffer()
//line app/vmselect/prometheus/series_response.qtpl:30
	WriteSeriesResponse(qb422016, metricNames, qt, qtDone)
//line app/vmselect/prometheus/series_response.qtpl:30
	qs422016 := string(qb422016.B)
//line app/vmselect/prometheus/series_response.qtpl:30
	qt422016.ReleaseByteBuffer(qb422016)
//line app/vmselect/prometheus/series_response.qtpl:30
	return qs422016
//line app/vmselect/prometheus/series_response.qtpl:30
}
